
local vec2 = require 'vec2'

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local Circle = {}
Circle.__index = Circle

local function new_circle_shape(radius)
	return setmetatable({
		bounding_radius = radius;
	}, Circle)
end

function Circle:support(dx,dy)
	return vec2.abs_scale(self.bounding_radius, dx,dy)
end

function Circle:project_onto_axis(axis_x,axis_y)
	return -self.bounding_radius, self.bounding_radius
end

function Circle:intersects_ray(x,y, dx,dy)
	local offset = vec2.dot(x, y, vec2.perp(dx, dy))
	if offset <= self.bounding_radius then
		local pos_along_ray = -vec2.dot(x, y, dx, dy)
		local bulge_at_offset = math.sqrt(self.bounding_radius*self.bounding_radius - offset*offset)
		if pos_along_ray >= bulge_at_offset then
			local distance = pos_along_ray - bulge_at_offset
			return distance, vec2.safe_normalize(x+dx*distance, y+dy*distance)
		end
	end
end

function Circle:debug_draw(style)
	love.graphics.circle(style, 0, 0, self.bounding_radius)
end

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local Capsule = {}
Capsule.__index = Capsule

local function new_capsule_shape(radius, separation_x, separation_y)
	local nx, ny = vec2.safe_norm(separation_x, separation_y)
	return setmetatable({
		_radius = radius,
		_separation_x = separation_x,
		_separation_y = separation_y,
		_norm_separation_x = nx,
		_norm_separation_y = ny,
		bounding_radius = radius + vec2.len(separation_x, separation_y),
	}, Capsule)
end

function Capsule:support(dx,dy)
	local circle_support_x, circle_support_y = vec2.abs_scale(self.bounding_radius, dx,dy)
	if vec2.dot(dx,dy, self._separation_x, self._separation_y) then
		return circle_support_x + self._separation_x, circle_support_y + self._separation_y
	else
		return circle_support_x - self._separation_x, circle_support_y - self._separation_y
	end
end

function Capsule:project_onto_axis(axis_x,axis_y)
	local sep = math.abs(vec2.dot(axis_x, axis_y, self._separation_x, self._separation_y))
	return -self._radius - sep, self._radius + sep
end

function Capsule:intersects_ray(x,y, dx,dy)
	error('Capsule:intersects_ray not yet implemented')
end

function Capsule:debug_draw(style)
	error('Capsule:debug_draw not yet implemented')
end

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local Polygon = {}
Polygon.__index = Polygon

local function new_polygon_shape(...)
	local vertices = ...
	if type(vertices) ~= 'table' then
		vertices = {...}
	end
	local vertices_len = #vertices
	local radius2 = 0
	local sat_normals = {}
	local x1, y1 = vertices[vertices_len-1], vertices[vertices_len]
	for i = 1, vertices_len, 2 do
		local x2, y2 = vertices[i], vertices[i+1]
		local nx, ny = vec2.normalize(y2 - y1, x1 - x2)
		sat_normals[i], sat_normals[i+1] = nx, ny
		local length2 = vec2.length2(x2, y2)
		radius2 = math.max(radius2, length2)
		x1, y1 = x2, y2
	end
	return setmetatable({
		_vertices = vertices;
		_vertices_len = vertices_len;
		_sat_normals = sat_normals;
		_sat_normals_len = #sat_normals;
		bounding_radius = math.sqrt(radius2);
	}, Polygon)
end

local function new_box_shape(x1, y1, x2, y2)
	return new_polygon_shape(x1,y1, x2,y1, x2,y2, x1,y2)
end

function Polygon:support(dx,dy)
	local maxdot = -math.huge
	local support_x, support_y
	for i = 1, self._vertices_len, 2 do
		local x, y = self._vertices[i], self._vertices[i+1]
		local dot = vec2.dot(dx,dy, x,y)
		if dot > maxdot then
			maxdot, support_x, support_y = dot, x, y
		end
	end
	return support_x, support_y
end

function Polygon:project_onto_axis(axis_x,axis_y)
	local min, max = math.huge, -math.huge
	for i = 1, self._vertices_len, 2 do
		local projection = vec2.dot(axis_x, axis_y, self._vertices[i], self._vertices[i+1])
		min = math.min(min, projection)
		max = math.max(max, projection)
	end
	return min, max
end

function Polygon:intersects_ray(x,y, dx,dy)
	local pdx, pdy = vec2.perp(dx, dy)
	local start_offset = vec2.dot(x, y, pdx, pdy)
	local x1, y1 = self._vertices[self._vertices_len-1], self._vertices[self._vertices_len]
	for i = 1, self._vertices_len, 2 do
		local x2, y2 = self._vertices[i], self._vertices[i+1]
		local v1_offset = vec2.dot(x1, y1, pdx, pdy)
		local v2_offset = vec2.dot(x2, y2, pdx, pdy)
		if v1_offset >= start_offset and v2_offset <= start_offset then
			local start_along_ray = vec2.dot(x, y, dx, dy)
			local v1_along_ray = vec2.dot(x1, y1, dx, dy)
			local v2_along_ray = vec2.dot(x2, y2, dx, dy)
			local intersection_along_ray = v1_along_ray
				+ ((start_offset - v1_offset) / (v2_offset - v1_offset))
				* (v2_along_ray - v1_along_ray)
			if intersection_along_ray >= start_along_ray then
				return intersection_along_ray - start_along_ray,
					self._sat_normals[i], self._sat_normals[i+1]
			end
		end
		x1, y1 = x2, y2
	end
end

function Polygon:debug_draw(style)
	love.graphics.polygon(style, unpack(self._vertices))
end

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local InfiniteBoundary = {}
InfiniteBoundary.__index = InfiniteBoundary

local function new_infinite_boundary_shape(norm_x, norm_y)
	return setmetatable({
		_norm_x = norm_x,
		_norm_y = norm_y,
		bounding_radius = math.huge,
	}, InfiniteBoundary)
end

function InfiniteBoundary:support(dx,dy)
	error("InfiniteBoundary:support() not implemented")
end

function InfiniteBoundary:project_onto_axis(axis_x,axis_y)
	local dot = vec2.dot(self._norm_x, self._norm_y, axis_x, axis_y)
	if dot > 0.9999 then
		return -math.huge, 0
	elseif dot < -0.9999 then
		return 0, math.huge
	end
	return -math.huge, math.huge
end

function InfiniteBoundary:intersects_ray(x,y, dx,dy)
	local perp_dist = vec2.dot(x, y, self._norm_x, self._norm_y)
	if perp_dist >= 0 then
		local dir_dot = vec2.dot(self._norm_x, self._norm_y, dx, dy)
		if dir_dot < 0 then
			return perp_dist / -dir_dot, self._norm_x, self._norm_y
		end
	end
end

function InfiniteBoundary:debug_draw(style)
	error('InfiniteBoundary:debug_draw not yet implemented')
end

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local function intersects_circle_circle(a, b, ax,ay,ac,as, bx,by,bc,bs)
	local total_radius = a.bounding_radius + b.bounding_radius
	local diff_x, diff_y = vec2.diff(ax, ay, bx, by)
	local distance2 = vec2.len2(diff_x, diff_y)
	if distance2 < total_radius * total_radius then
		local distance = math.sqrt(distance2)
		if distance == 0 then
			return true, total_radius, 0
		end
		local intersection = distance - total_radius
		return true, diff_x / distance * intersection, diff_y / distance * intersection
	end
	return false
end

local function try_sat_projection(min_axis_x,min_axis_y, min_translation, a, b, ax,ay,ac,as, bx,by,bc,bs, axis_x,axis_y)
	local offset_a = vec2.dot(ax, ay, axis_x, axis_y)
	local offset_b = vec2.dot(bx, by, axis_x, axis_y)
	-- the outer vec2.add isn't being used on a vector, but for convenience on two separate numbers
	local proj_a1, proj_a2 = vec2.add(offset_a, offset_a, a:project_onto_axis(vec2.to_x_axis(ac, as, axis_x, axis_y)))
	local proj_b1, proj_b2 = vec2.add(offset_b, offset_b, b:project_onto_axis(vec2.to_x_axis(bc, bs, axis_x, axis_y)))
	if not (proj_a1 < proj_b2 and proj_b1 < proj_a2) then
		return
	end
	local trans1, trans2 = proj_b1 - proj_a2, proj_b2 - proj_a1
	local translation = trans1*trans1 > trans2*trans2 and trans2 or trans1
	if translation*translation < min_translation*min_translation then
		return axis_x, axis_y, translation
	else
		return min_axis_x, min_axis_y, min_translation
	end
end

local function polygon_get_axis_to(p, circle_x,circle_y, px,py,pc,ps)
	local min, axis_x, axis_y
	for i = 1, p._vertices_len, 2 do
		local x, y = vec2.add(px, py, vec2.from_x_axis(pc, ps, p._vertices[i], p._vertices[i+1]))
		local ax, ay = vec2.diff(x, y, circle_x, circle_y)
		local length2 = vec2.len2(ax, ay)
		if not min or length2 < min then
			min, axis_x, axis_y = length2, ax, ay
		end
	end
	return vec2.safe_norm(axis_x, axis_y)
end

local function bounding_radius_check(a, b, ax,ay, bx,by)
	local total_radius = a.bounding_radius + b.bounding_radius
	return vec2.len2(vec2.diff(ax, ay, bx, by)) <= total_radius * total_radius
end

local function intersects_polygon_polygon(a, b, ax,ay,ac,as, bx,by,bc,bs)
	if not bounding_radius_check(a, b, ax,ay, bx,by) then return false end
	local min_axis_x, min_axis_y, min_translation= nil, nil, math.huge
	
	for i = 1, a._sat_normals_len, 2 do
		min_axis_x, min_axis_y, min_translation = try_sat_projection(
			min_axis_x, min_axis_y, min_translation,
			a, b, ax,ay,ac,as, bx,by,bc,bs,
			vec2.from_x_axis(ac, as, a._sat_normals[i], a._sat_normals[i+1]))
		if not min_axis_x then return false end
	end
	
	for i = 1, b._sat_normals_len, 2 do
		min_axis_x, min_axis_y, min_translation = try_sat_projection(
			min_axis_x, min_axis_y, min_translation,
			a, b, ax,ay,ac,as, bx,by,bc,bs,
			vec2.from_x_axis(bc, bs, b._sat_normals[i], b._sat_normals[i+1]))
		if not min_axis_x then return false end
	end
	
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

local function intersects_polygon_circle(a, b, ax,ay,ac,as, bx,by,bc,bs)
	if not bounding_radius_check(a, b, ax,ay, bx,by) then return false end
	local min_axis_x, min_axis_y, min_translation = nil, nil, math.huge
	
	for i = 1, a._sat_normals_len, 2 do
		min_axis_x, min_axis_y, min_translation = try_sat_projection(
			min_axis_x, min_axis_y, min_translation,
			a, b, ax,ay,ac,as, bx,by,bc,bs,
			vec2.from_x_axis(ac, as, a._sat_normals[i], a._sat_normals[i+1]))
		if not min_axis_x then return false end
	end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		polygon_get_axis_to(a, bx, by, ax,ay,ac,as))
	if not min_axis_x then return false end
	
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

local function intersects_polygon_capsule(a, b, ax,ay,ac,as, bx,by,bc,bs)
	if not bounding_radius_check(a, b, ax,ay, bx,by) then return false end
	local min_axis_x, min_axis_y, min_translation = nil, nil, math.huge
	
	for i = 1, a._sat_normals_len, 2 do
		min_axis_x, min_axis_y, min_translation = try_sat_projection(
			min_axis_x, min_axis_y, min_translation,
			a, b, ax,ay,ac,as, bx,by,bc,bs,
			vec2.from_x_axis(ac, as, a._sat_normals[i], a._sat_normals[i+1]))
		if not min_axis_x then return false end
	end
	
	local px, py = vec2.add(bx, by, vec2.from_x_axis(bc, bs, b._separation_x, b._separation_y))
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		polygon_get_axis_to(a, px, py, ax,ay,ac,as))
	if not min_axis_x then return false end
	
	local px, py = vec2.sub(bx, by, vec2.from_x_axis(bc, bs, b._separation_x, b._separation_y))
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		polygon_get_axis_to(a, px, py, ax,ay,ac,as))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.from_x_axis(bc, bs, vec2.perp(b._norm_separation_x, b._norm_separation_y)))
	if not min_axis_x then return false end
	
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

local function intersects_circle_capsule(a, b, ax,ay,ac,as, bx,by,bc,bs)
	if not bounding_radius_check(a, b, ax,ay, bx,by) then return false end
	local min_axis_x, min_axis_y, min_translation = nil, nil, math.huge
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(ax, ay, vec2.add(bx, by, vec2.from_x_axis(bc, bs, b._separation_x, b._separation_y)))))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(ax, ay, vec2.add(bx, by, vec2.from_x_axis(bc, bs, -b._separation_x, -b._separation_y)))))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.from_x_axis(bc, bs, vec2.perp(b._norm_separation_x, b._norm_separation_y)))
	if not min_axis_x then return false end
	
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

local function intersects_capsule_capsule(a, b, ax,ay,ac,as, bx,by,bc,bs)
	if not bounding_radius_check(a, b, ax,ay, bx,by) then return false end
	local min_axis_x, min_axis_y, min_translation = nil, nil, math.huge
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.from_x_axis(ac, as, vec2.perp(a._norm_separation_x, a._norm_separation_y)))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.from_x_axis(bc, bs, vec2.perp(b._norm_separation_x, b._norm_separation_y)))
	if not min_axis_x then return false end
	
	local asx, asy = vec2.from_x_axis(ac, as, a._separation_x, a._separation_y)
	local bsx, bsy = vec2.from_x_axis(bc, bs, b._separation_x, b._separation_y)
	local a1x, a1y = vec2.add(ax, ay, asx, asy)
	local a2x, a2y = vec2.sub(ax, ay, asx, asy)
	local b1x, b1y = vec2.add(bx, by, bsx, bsy)
	local b2x, b2y = vec2.sub(bx, by, bsx, bsy)

	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(a1x, a1y, b1x, b1y)))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(a1x, a1y, b2x, b2y)))
	if not min_axis_x then return false end

	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(a2x, a2y, b1x, b1y)))
	if not min_axis_x then return false end
	
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.safe_norm(vec2.diff(a2x, a2y, b2x, b2y)))
	if not min_axis_x then return false end
	
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

local function intersects_infinite_boundary_any(a, b, ax,ay,ac,as, bx,by,bc,bs)
	local min_axis_x, min_axis_y, min_translation = nil, nil, math.huge
	min_axis_x, min_axis_y, min_translation = try_sat_projection(
		min_axis_x, min_axis_y, min_translation,
		a, b, ax,ay,ac,as, bx,by,bc,bs,
		vec2.from_x_axis(ac, as, a._norm_x, a._norm_y))
	if not min_axis_x then return false end
	return true, min_axis_x * min_translation, min_axis_y * min_translation
end

------------------------------------------------------------------------------
------------------------------------------------------------------------------

local function invert_intersection_function(f)
	return function (a, b, ax,ay,ac,as, bx,by,bc,bs)
		local hit, dx, dy = f(b, a, bx,by,bc,bs, ax,ay,ac,as)
		if hit then
			return true, -dx, -dy
		end
		return false
	end
end

local intersection_function_table = {
	[Circle] = {
		[Circle] = intersects_circle_circle;
		[Polygon] = invert_intersection_function(intersects_polygon_circle);
		[Capsule] = intersects_circle_capsule;
		[InfiniteBoundary] = invert_intersection_function(intersects_infinite_boundary_any),
	};
	[Polygon] = {
		[Circle] = intersects_polygon_circle;
		[Polygon] = intersects_polygon_polygon;
		[Capsule] = intersects_polygon_capsule;
		[InfiniteBoundary] = invert_intersection_function(intersects_infinite_boundary_any),
	};
	[Capsule] = {
		[Circle] = invert_intersection_function(intersects_circle_capsule);
		[Polygon] = invert_intersection_function(intersects_polygon_capsule);
		[Capsule] = intersects_capsule_capsule,
		[InfiniteBoundary] = invert_intersection_function(intersects_infinite_boundary_any),
	};
	[InfiniteBoundary] = {
		[Circle] = intersects_infinite_boundary_any,
		[Polygon] = intersects_infinite_boundary_any,
		[Capsule] = intersects_infinite_boundary_any,
		[InfiniteBoundary] = function() return false end,
	};
}

local function are_intersecting(a, b, ax,ay,ac,as, bx,by,bc,bs)
	return intersection_function_table[getmetatable(a)][getmetatable(b)](
		a,b, ax,ay,ac,as, bx,by,bc,bs)
end

Circle.intersects = are_intersecting
Polygon.intersects = are_intersecting
Capsule.intersects = are_intersecting
InfiniteBoundary.intersects = are_intersecting

------------------------------------------------------------------------------
------------------------------------------------------------------------------

return {
	new_circle_shape = new_circle_shape;
	new_capsule_shape = new_capsule_shape;
	new_polygon_shape = new_polygon_shape;
	new_box_shape = new_box_shape;
	new_infinite_boundary_shape = new_infinite_boundary_shape;
}

